@@ -31,10 +31,12 @@ pub mod rollsum;
 pub mod sendlog;
 pub mod server;
 pub mod sodium;
+pub mod xglobset;
 pub mod xid;
 pub mod xtar;
 
-use std::collections::BTreeMap;
+use std::collections::{BTreeMap, HashMap};
+use std::fmt::Write as FmtWrite;
 use std::io::{BufRead, Read, Write};
 use std::os::unix::io::{AsRawFd, FromRawFd, IntoRawFd};
 use std::path::{Path, PathBuf};
@@ -62,7 +64,12 @@ fn cache_dir() -> Result<PathBuf, anyhow::Error> {
 }
 
 fn print_help_and_exit(subcommand: &str, opts: &getopts::Options) {
-    let brief = match subcommand {
+    let subcommand_no_alias = match subcommand {
+        "remove" => "rm",
+        subcommand => subcommand,
+    };
+
+    let mut brief = match subcommand_no_alias {
         "init" => include_str!("../doc/cli/init.txt"),
         "help" => include_str!("../doc/cli/help.txt"),
         "new-key" => include_str!("../doc/cli/new-key.txt"),
@@ -73,15 +80,29 @@ fn print_help_and_exit(subcommand: &str, opts: &getopts::Options) {
         "diff" => include_str!("../doc/cli/diff.txt"),
         "get" => include_str!("../doc/cli/get.txt"),
         "restore" => include_str!("../doc/cli/restore.txt"),
-        "rm" | "remove" => include_str!("../doc/cli/rm.txt"),
+        "rm" => include_str!("../doc/cli/rm.txt"),
         "recover-removed" => include_str!("../doc/cli/recover-removed.txt"),
         "gc" => include_str!("../doc/cli/gc.txt"),
         "serve" => include_str!("../doc/cli/serve.txt"),
+        "sync" => include_str!("../doc/cli/sync.txt"),
         "version" => include_str!("../doc/cli/version.txt"),
+        "exec-with-locks" => include_str!("../doc/cli/exec-with-locks.txt"),
         "put-benchmark" => "put-benchmark tool.",
         _ => panic!(),
-    };
-    let _ = std::io::stdout().write_all(opts.usage(brief).as_bytes());
+    }
+    .to_string();
+
+    writeln!(brief, "\n\nOnline Manual:").unwrap();
+    write!(
+        brief,
+        "  https://bupstash.io/doc/{}/man/bupstash-{}.html",
+        env!("CARGO_PKG_VERSION"),
+        subcommand_no_alias
+    )
+    .unwrap();
+
+    let _ = std::io::stdout().write_all(opts.usage(&brief).as_bytes());
+
     std::process::exit(0);
 }
 
@@ -253,21 +274,31 @@ fn cli_to_query_cache(matches: &getopts::Matches) -> Result<querycache::QueryCac
     }
 }
 
-fn cli_to_id_and_query(
+fn cli_to_id_and_opt_query(
     matches: &getopts::Matches,
-) -> Result<(Option<xid::Xid>, query::Query), anyhow::Error> {
-    let query: query::Query = if !matches.free.is_empty() {
+) -> Result<(Option<xid::Xid>, Option<query::Query>), anyhow::Error> {
+    if !matches.free.is_empty() {
         match query::parse(&matches.free.join("•")) {
-            Ok(query) => query,
+            Ok(query) => Ok((query::get_id_query(&query), Some(query))),
             Err(e) => {
                 query::report_parse_error(e);
                 anyhow::bail!("query parse error");
             }
         }
+    } else {
+        Ok((None, None))
+    }
+}
+
+fn cli_to_id_and_query(
+    matches: &getopts::Matches,
+) -> Result<(Option<xid::Xid>, query::Query), anyhow::Error> {
+    let (id, query) = cli_to_id_and_opt_query(matches)?;
+    let query = if let Some(query) = query {
+        query
     } else {
         anyhow::bail!("you must specify a query");
     };
-    let id = query::get_id_query(&query);
     Ok((id, query))
 }
 
@@ -304,15 +335,33 @@ impl Drop for ServeProcess {
     }
 }
 
+#[derive(Clone, Copy)]
+struct ServeProcessCliOpts<'a, 'b, 'c> {
+    repository_arg: &'a str,
+    repository_env_var: &'b str,
+    repository_command_env_var: &'c str,
+}
+
+impl<'a, 'b, 'c> Default for ServeProcessCliOpts<'a, 'b, 'c> {
+    fn default() -> Self {
+        ServeProcessCliOpts {
+            repository_arg: "repository",
+            repository_env_var: "BUPSTASH_REPOSITORY",
+            repository_command_env_var: "BUPSTASH_REPOSITORY_COMMAND",
+        }
+    }
+}
+
 fn cli_to_serve_process(
     matches: &getopts::Matches,
     progress: &indicatif::ProgressBar,
+    opts: ServeProcessCliOpts,
 ) -> Result<ServeProcess, anyhow::Error> {
     let mut serve_cmd_args = {
-        let repo = if matches.opt_present("repository") {
-            Some(matches.opt_str("repository").unwrap())
+        let repo = if matches.opt_present(opts.repository_arg) {
+            Some(matches.opt_str(opts.repository_arg).unwrap())
         } else {
-            std::env::var_os("BUPSTASH_REPOSITORY").map(|r| r.into_string().unwrap())
+            std::env::var_os(opts.repository_env_var).map(|r| r.into_string().unwrap())
         };
 
         match repo {
@@ -349,21 +398,27 @@ fn cli_to_serve_process(
                 }
             }
             None => {
-                if let Some(connect_cmd) = std::env::var_os("BUPSTASH_REPOSITORY_COMMAND") {
+                if let Some(connect_cmd) = std::env::var_os(opts.repository_command_env_var) {
                     match shlex::split(&connect_cmd.into_string().unwrap()) {
                         Some(args) => {
                             if args.is_empty() {
                                 anyhow::bail!(
-                                    "BUPSTASH_REPOSITORY_COMMAND should have at least one element"
+                                    "{} should have at least one element",
+                                    opts.repository_command_env_var,
                                 );
                             }
                             args
                         }
-                        None => anyhow::bail!("unable to parse BUPSTASH_REPOSITORY_COMMAND"),
+                        None => {
+                            anyhow::bail!("unable to parse {}", opts.repository_command_env_var)
+                        }
                     }
                 } else {
                     anyhow::bail!(
-                        "please set --repository, BUPSTASH_REPOSITORY or BUPSTASH_REPOSITORY_COMMAND"
+                        "please set --{}, {} or {}",
+                        opts.repository_arg,
+                        opts.repository_env_var,
+                        opts.repository_command_env_var,
                     );
                 }
             }
@@ -418,13 +473,14 @@ fn cli_to_serve_process(
 fn cli_to_opened_serve_process(
     matches: &getopts::Matches,
     progress: &indicatif::ProgressBar,
+    serve_process_opts: ServeProcessCliOpts,
     open_mode: protocol::OpenMode,
 ) -> Result<ServeProcess, anyhow::Error> {
     let mut retry_delay_secs = 2;
     let mut retry_count: u64 = 0;
 
     loop {
-        let mut remote = cli_to_serve_process(matches, progress)?;
+        let mut remote = cli_to_serve_process(matches, progress, serve_process_opts)?;
         // Temporary borrow of the stdin/stdout so we can handle
         // server retry back pressure if the server implements this.
         let mut proc_stdin = remote.proc.stdin.take().unwrap();
@@ -536,7 +592,7 @@ fn init_main(args: Vec<String>) -> Result<(), anyhow::Error> {
         indicatif::ProgressStyle::default_spinner().template("[{elapsed_precise}] {wide_msg}"),
     );
 
-    let mut serve_proc = cli_to_serve_process(&matches, &progress)?;
+    let mut serve_proc = cli_to_serve_process(&matches, &progress, ServeProcessCliOpts::default())?;
     let mut serve_out = serve_proc.proc.stdout.as_mut().unwrap();
     let mut serve_in = serve_proc.proc.stdin.as_mut().unwrap();
 
@@ -628,8 +684,12 @@ fn list_main(args: Vec<String>) -> Result<(), anyhow::Error> {
 
     let mut query_cache = cli_to_query_cache(&matches)?;
 
-    let mut serve_proc =
-        cli_to_opened_serve_process(&matches, &progress, protocol::OpenMode::Read)?;
+    let mut serve_proc = cli_to_opened_serve_process(
+        &matches,
+        &progress,
+        ServeProcessCliOpts::default(),
+        protocol::OpenMode::Read,
+    )?;
     let mut serve_out = serve_proc.proc.stdout.as_mut().unwrap();
     let mut serve_in = serve_proc.proc.stdin.as_mut().unwrap();
 
@@ -666,7 +726,7 @@ fn list_main(args: Vec<String>) -> Result<(), anyhow::Error> {
                             out,
                             "{}=\"{}\"",
                             k,
-                            v.replace("\\", "\\\\").replace("\"", "\\\"")
+                            v.replace('\\', "\\\\").replace('\"', "\\\"")
                         )?;
                     }
                     writeln!(out)?;
@@ -789,7 +849,12 @@ fn put_main(args: Vec<String>) -> Result<(), anyhow::Error> {
         "Primary or put key to encrypt data with.",
         "PATH",
     );
-    opts.optflag("", "no-compression", "Disable data compression.");
+    opts.optopt(
+        "",
+        "compression",
+        "Compression algorithm, one of 'none', 'lz4' or 'zstd'[:$level]. Defaults to 'zstd:3'.",
+        "COMPRESS",
+    );
     opts.optflag("", "no-default-tags", "Disable the default tag(s) 'name'.");
     opts.optflag(
         "v",
@@ -825,7 +890,11 @@ fn put_main(args: Vec<String>) -> Result<(), anyhow::Error> {
         "no-send-log",
         "Disable logging of previously sent data, implies --no-stat-caching.",
     );
-    opts.optflag("", "xattrs", "Save directory entry xattrs.");
+    opts.optflag(
+        "",
+        "xattrs",
+        "Save directory entry xattrs (at some performance cost).",
+    );
     opts.optopt(
         "",
         "send-log",
@@ -835,9 +904,19 @@ fn put_main(args: Vec<String>) -> Result<(), anyhow::Error> {
     opts.optmulti(
         "",
         "exclude",
-        "Exclude directory entries matching the given glob pattern when saving a directory, may be passed multiple times.",
+        "Exclude directory entries matching the given glob pattern when saving a directory, may be passed multiple times.\
+        Paths are absolute as they are currently mounted, and must start with a slash and not end on one even for directories.\
+        Patterns without any slashes will match any file (`--exclude foo` is equivalent to `--exclude '/**/foo'`).",
         "PATTERN",
     );
+    opts.optmulti(
+        "",
+        "exclude-if-present",
+        "Exclude a directory's content if it contains a file with the given name. May be passed multiple times.
+  This will still backup the folder itself, containing the marker file. Common marker file names are `CACHEDIR.TAG`, `.backupexclude`
+  or `.no-backup`.",
+        "FILENAME",
+    );
     opts.optflag(
         "",
         "one-file-system",
@@ -879,10 +958,11 @@ fn put_main(args: Vec<String>) -> Result<(), anyhow::Error> {
 
     let want_xattrs = matches.opt_present("xattrs");
 
-    let compression = if matches.opt_present("no-compression") {
-        compression::Scheme::None
-    } else {
-        compression::Scheme::Lz4
+    let compression = {
+        let scheme = matches
+            .opt_str("compression")
+            .unwrap_or_else(|| "zstd:3".to_string());
+        compression::parse_scheme(&scheme)?
     };
 
     let print_stats = matches.opt_present("print-stats") || matches.opt_present("verbose");
@@ -892,13 +972,63 @@ fn put_main(args: Vec<String>) -> Result<(), anyhow::Error> {
     let use_stat_cache =
         !(matches.opt_present("no-stat-caching") || matches.opt_present("no-send-log"));
 
-    let mut exclusions = Vec::new();
-    for e in matches.opt_strs("exclude") {
-        match glob::Pattern::new(&e) {
-            Ok(pattern) => exclusions.push(pattern),
-            Err(err) => anyhow::bail!("--exclude option {:?} is not a valid glob: {}", e, err),
+    let mut exclusions = globset::GlobSetBuilder::new();
+    for mut e in matches.opt_strs("exclude") {
+        /* Sanity checks. Beware, the order of the checks matters */
+
+        /* An exclude path ending on / won't match anything. */
+        if e.ends_with('/') {
+            anyhow::bail!(
+                "--exclude option '{}' ends with '/', so it won't match anything",
+                e
+            );
         }
+
+        /* This check is technically redundant, but it gives a nicer error message. */
+        if e.starts_with("./") || e.starts_with("../") {
+            anyhow::bail!("relative paths are not allowed in --exclude patterns");
+        }
+
+        /* Start with a / to match a path, and leave out slashes to match any file. */
+        if e.starts_with('/') || e.starts_with("**/") {
+            /* pass */
+        } else if e.contains('/') && !e.contains('[') {
+            /* This is just to help the user with a nicer error message, we do
+             * not take this branch if the pattern contains an escape character. */
+            anyhow::bail!(
+                "--exclude option '{}' contains '/' so must be absolute to match anything",
+                e
+            );
+        } else {
+            /* Just a file name */
+            e = format!("**/{}", e);
+        }
+
+        /* Check for unnormalized segments, as they too won't match anything. Skip this
+         * check if the pattern contains the escape character. */
+        if (e.contains("/./") || e.contains("/../") || e.contains("//")) && !e.contains('[') {
+            anyhow::bail!(
+                "--exclude option '{}' must be normalized (no '.', '..' or '//' path segments)",
+                e
+            );
+        }
+
+        let mut pattern = globset::GlobBuilder::new(&e);
+        /* For some reason, the default doesn't give us the common behavior one would expect. */
+        pattern.literal_separator(true);
+        pattern.backslash_escape(true);
+        let pattern = pattern.build().map_err(|err| {
+            anyhow::format_err!("--exclude option '{}' is not a valid glob: {}", e, err)
+        })?;
+        exclusions.add(pattern);
     }
+    let exclusions = exclusions.build()?;
+
+    let exclusion_markers = matches
+        .opt_strs("exclude-if-present")
+        .drain(..)
+        .map(std::ffi::OsString::from)
+        .collect();
 
     let one_file_system = matches.opt_present("one-file-system");
 
@@ -1031,6 +1161,7 @@ fn put_main(args: Vec<String>) -> Result<(), anyhow::Error> {
                 data_source = client::DataSource::Filesystem {
                     paths: vec![input_path],
                     exclusions,
+                    exclusion_markers,
                 };
             } else if md.is_file() {
                 if default_tags && !tags.contains_key("name") {
@@ -1071,6 +1202,7 @@ fn put_main(args: Vec<String>) -> Result<(), anyhow::Error> {
         data_source = client::DataSource::Filesystem {
             paths: absolute_paths,
             exclusions,
+            exclusion_markers,
         };
     };
 
@@ -1098,8 +1230,12 @@ fn put_main(args: Vec<String>) -> Result<(), anyhow::Error> {
         }
     };
 
-    let mut serve_proc =
-        cli_to_opened_serve_process(&matches, &progress, protocol::OpenMode::ReadWrite)?;
+    let mut serve_proc = cli_to_opened_serve_process(
+        &matches,
+        &progress,
+        ServeProcessCliOpts::default(),
+        protocol::OpenMode::ReadWrite,
+    )?;
     let mut serve_out = serve_proc.proc.stdout.as_mut().unwrap();
     let mut serve_in = serve_proc.proc.stdin.as_mut().unwrap();
 
@@ -1219,8 +1355,12 @@ fn get_main(args: Vec<String>) -> Result<(), anyhow::Error> {
     );
 
     let (id, query) = cli_to_id_and_query(&matches)?;
-    let mut serve_proc =
-        cli_to_opened_serve_process(&matches, &progress, protocol::OpenMode::Read)?;
+    let mut serve_proc = cli_to_opened_serve_process(
+        &matches,
+        &progress,
+        ServeProcessCliOpts::default(),
+        protocol::OpenMode::Read,
+    )?;
     let mut serve_out = serve_proc.proc.stdout.as_mut().unwrap();
     let mut serve_in = serve_proc.proc.stdin.as_mut().unwrap();
 
@@ -1301,8 +1441,8 @@ fn get_main(args: Vec<String>) -> Result<(), anyhow::Error> {
         progress.set_message("picking content...");
 
         if let Some(ref index) = get_index {
-            let (pick_index, pick_data_map) =
-                index::pick(&matches.opt_str("pick").unwrap(), index)?;
+            let pick_path: PathBuf = matches.opt_str("pick").unwrap().into();
+            let (pick_index, pick_data_map) = index::pick(&pick_path, index)?;
             get_index = pick_index;
             get_data_map = Some(pick_data_map);
         } else {
@@ -1404,8 +1544,12 @@ fn list_contents_main(args: Vec<String>) -> Result<(), anyhow::Error> {
     );
 
     let (id, query) = cli_to_id_and_query(&matches)?;
-    let mut serve_proc =
-        cli_to_opened_serve_process(&matches, &progress, protocol::OpenMode::Read)?;
+    let mut serve_proc = cli_to_opened_serve_process(
+        &matches,
+        &progress,
+        ServeProcessCliOpts::default(),
+        protocol::OpenMode::Read,
+    )?;
     let mut serve_out = serve_proc.proc.stdout.as_mut().unwrap();
     let mut serve_in = serve_proc.proc.stdin.as_mut().unwrap();
 
@@ -1485,8 +1629,8 @@ fn list_contents_main(args: Vec<String>) -> Result<(), anyhow::Error> {
 
     if matches.opt_present("pick") {
         progress.set_message("picking content...");
-        content_index =
-            index::pick_dir_without_data(&matches.opt_str("pick").unwrap(), &content_index)?;
+        let pick_path: PathBuf = matches.opt_str("pick").unwrap().into();
+        content_index = index::pick_dir_without_data(&pick_path, &content_index)?;
     }
 
     client::hangup(&mut serve_in)?;
@@ -1664,8 +1808,12 @@ fn diff_main(args: Vec<String>) -> Result<(), anyhow::Error> {
         indicatif::ProgressStyle::default_spinner().template("[{elapsed_precise}] {wide_msg}"),
     );
 
-    let mut serve_proc =
-        cli_to_opened_serve_process(&matches, &progress, protocol::OpenMode::Read)?;
+    let mut serve_proc = cli_to_opened_serve_process(
+        &matches,
+        &progress,
+        ServeProcessCliOpts::default(),
+        protocol::OpenMode::Read,
+    )?;
     let mut serve_out = serve_proc.proc.stdout.as_mut().unwrap();
     let mut serve_in = serve_proc.proc.stdin.as_mut().unwrap();
 
@@ -1679,8 +1827,10 @@ fn diff_main(args: Vec<String>) -> Result<(), anyhow::Error> {
             for indexed_dir in indexer::FsIndexer::new(
                 &paths,
                 indexer::FsIndexerOptions {
-                    exclusions: vec![],
+                    exclusions: globset::GlobSet::empty(),
+                    exclusion_markers: std::collections::HashSet::new(),
                     want_xattrs: matches.opt_present("xattrs"),
+                    want_sparseness: false,
                     want_hash: true,
                     one_file_system: false,
                 },
@@ -1784,8 +1934,8 @@ fn diff_main(args: Vec<String>) -> Result<(), anyhow::Error> {
     for (i, pick_opt) in ["left-pick", "right-pick"].iter().enumerate() {
         if matches.opt_present(pick_opt) {
             progress.set_message("picking content...");
-            to_diff[i] =
-                index::pick_dir_without_data(&matches.opt_str(pick_opt).unwrap(), &to_diff[i])?;
+            let pick_path: PathBuf = matches.opt_str(pick_opt).unwrap().into();
+            to_diff[i] = index::pick_dir_without_data(&pick_path, &to_diff[i])?;
         }
     }
 
@@ -1895,8 +2045,12 @@ fn remove_main(args: Vec<String>) -> Result<(), anyhow::Error> {
             };
         }
 
-        let mut serve_proc =
-            cli_to_opened_serve_process(&matches, &progress, protocol::OpenMode::ReadWrite)?;
+        let mut serve_proc = cli_to_opened_serve_process(
+            &matches,
+            &progress,
+            ServeProcessCliOpts::default(),
+            protocol::OpenMode::ReadWrite,
+        )?;
         let mut serve_out = serve_proc.proc.stdout.as_mut().unwrap();
         let mut serve_in = serve_proc.proc.stdin.as_mut().unwrap();
 
@@ -1904,8 +2058,12 @@ fn remove_main(args: Vec<String>) -> Result<(), anyhow::Error> {
         client::hangup(&mut serve_in)?;
         serve_proc.wait()?;
     } else {
-        let mut serve_proc =
-            cli_to_opened_serve_process(&matches, &progress, protocol::OpenMode::ReadWrite)?;
+        let mut serve_proc = cli_to_opened_serve_process(
+            &matches,
+            &progress,
+            ServeProcessCliOpts::default(),
+            protocol::OpenMode::ReadWrite,
+        )?;
         let mut serve_out = serve_proc.proc.stdout.as_mut().unwrap();
         let mut serve_in = serve_proc.proc.stdin.as_mut().unwrap();
 
@@ -1995,6 +2153,160 @@ fn remove_main(args: Vec<String>) -> Result<(), anyhow::Error> {
     Ok(())
 }
 
+fn sync_main(args: Vec<String>) -> Result<(), anyhow::Error> {
+    let mut opts = default_cli_opts();
+    repo_cli_opts(&mut opts);
+    query_cli_opts(&mut opts);
+
+    opts.optopt("k", "key", "Key to decrypt metadata with.", "PATH");
+
+    opts.optflag(
+        "",
+        "ids-from-stdin",
+        "Sync items with IDs read from stdin, one per line, instead of executing a query.",
+    );
+
+    opts.optopt(
+        "",
+        "to",
+        "Repository to sync items to, if prefixed with ssh:// implies ssh access. \
+         Defaults to BUPSTASH_TO_REPOSITORY if not set. \
+         See the manual for additional ways to connect to the repository.",
+        "REPO",
+    );
+
+    let matches = parse_cli_opts(opts, &args[..]);
+
+    let progress = cli_to_progress_bar(
+        &matches,
+        indicatif::ProgressStyle::default_spinner().template("[{elapsed_precise}] {wide_msg}"),
+    );
+
+    let mut ids = Vec::new();
+
+    let mut source_serve_proc = cli_to_opened_serve_process(
+        &matches,
+        &progress,
+        ServeProcessCliOpts::default(),
+        protocol::OpenMode::Read,
+    )?;
+    let mut source_serve_out = source_serve_proc.proc.stdout.as_mut().unwrap();
+    let mut source_serve_in = source_serve_proc.proc.stdin.as_mut().unwrap();
+
+    let mut ids_to_metadata: Option<HashMap<xid::Xid, oplog::VersionedItemMetadata>> = None;
+
+    if matches.opt_present("ids-from-stdin") {
+        for l in std::io::stdin().lock().lines() {
+            let l = l?;
+            if l.is_empty() {
+                continue;
+            }
+            match xid::Xid::parse(&l) {
+                Ok(id) => ids.push(id),
+                Err(err) => anyhow::bail!("error id parsing {:?}: {}", l, err),
+            };
+        }
+    } else {
+        match cli_to_id_and_opt_query(&matches)? {
+            (Some(id), _) => ids.push(id),
+            (_, query) => {
+                let mut query_cache = cli_to_query_cache(&matches)?;
+
+                // Only sync the client if we have a non id query.
+                client::sync_query_cache(
+                    progress.clone(),
+                    &mut query_cache,
+                    &mut source_serve_out,
+                    &mut source_serve_in,
+                )?;
+
+                let (primary_key_id, metadata_dctx) = match cli_to_opt_key(&matches)? {
+                    Some(key) => {
+                        if !key.is_list_key() {
+                            anyhow::bail!("only primary keys and sub keys created with '--list' can be used for listing")
+                        }
+
+                        let primary_key_id = key.primary_key_id();
+                        let metadata_dctx = match key {
+                            keys::Key::PrimaryKeyV1(k) => {
+                                crypto::DecryptionContext::new(k.metadata_sk, k.metadata_psk)
+                            }
+                            keys::Key::SubKeyV1(k) => crypto::DecryptionContext::new(
+                                k.metadata_sk.unwrap(),
+                                k.metadata_psk.unwrap(),
+                            ),
+                        };
+
+                        (Some(primary_key_id), Some(metadata_dctx))
+                    }
+                    None => (None, None),
+                };
+
+                ids_to_metadata = Some(HashMap::new());
+
+                let mut on_match =
+                    |item_id: xid::Xid,
+                     _tags: &std::collections::BTreeMap<String, String>,
+                     metadata: &oplog::VersionedItemMetadata,
+                     _secret_metadata: Option<&oplog::DecryptedItemMetadata>| {
+                        ids.push(item_id);
+                        ids_to_metadata
+                            .as_mut()
+                            .unwrap()
+                            .insert(item_id, metadata.clone());
+                        Ok(())
+                    };
+
+                let mut tx = query_cache.transaction()?;
+                tx.list(
+                    querycache::ListOptions {
+                        primary_key_id,
+                        metadata_dctx,
+                        list_encrypted: matches.opt_present("query-encrypted") || query.is_none(),
+                        utc_timestamps: matches.opt_present("utc-timestamps"),
+                        query,
+                        now: chrono::Utc::now(),
+                    },
+                    &mut on_match,
+                )?;
+            }
+        };
+    };
+
+    let mut dest_serve_proc = cli_to_opened_serve_process(
+        &matches,
+        &progress,
+        ServeProcessCliOpts {
+            repository_arg: "to",
+            repository_env_var: "BUPSTASH_TO_REPOSITORY",
+            repository_command_env_var: "BUPSTASH_TO_REPOSITORY_COMMAND",
+        },
+        protocol::OpenMode::ReadWrite,
+    )?;
+    let mut dest_serve_out = dest_serve_proc.proc.stdout.as_mut().unwrap();
+    let mut dest_serve_in = dest_serve_proc.proc.stdin.as_mut().unwrap();
+
+    client::repo_sync(
+        &progress,
+        ids,
+        ids_to_metadata,
+        &mut source_serve_out,
+        &mut source_serve_in,
+        &mut dest_serve_out,
+        &mut dest_serve_in,
+    )?;
+
+    client::hangup(&mut source_serve_in)?;
+    source_serve_proc.wait()?;
+
+    client::hangup(&mut dest_serve_in)?;
+    dest_serve_proc.wait()?;
+
+    progress.finish_and_clear();
+
+    Ok(())
+}
+
 fn gc_main(args: Vec<String>) -> Result<(), anyhow::Error> {
     let mut opts = default_cli_opts();
     opts.optflag("", "no-progress", "Suppress progress indicators.");
@@ -2008,7 +2320,12 @@ fn gc_main(args: Vec<String>) -> Result<(), anyhow::Error> {
         indicatif::ProgressStyle::default_spinner().template("[{elapsed_precise}] {wide_msg}"),
     );
 
-    let mut serve_proc = cli_to_opened_serve_process(&matches, &progress, protocol::OpenMode::Gc)?;
+    let mut serve_proc = cli_to_opened_serve_process(
+        &matches,
+        &progress,
+        ServeProcessCliOpts::default(),
+        protocol::OpenMode::Gc,
+    )?;
     let mut serve_out = serve_proc.proc.stdout.as_mut().unwrap();
     let mut serve_in = serve_proc.proc.stdin.as_mut().unwrap();
 
@@ -2050,8 +2367,12 @@ fn recover_removed(args: Vec<String>) -> Result<(), anyhow::Error> {
         indicatif::ProgressStyle::default_spinner().template("[{elapsed_precise}] {wide_msg}"),
     );
 
-    let mut serve_proc =
-        cli_to_opened_serve_process(&matches, &progress, protocol::OpenMode::ReadWrite)?;
+    let mut serve_proc = cli_to_opened_serve_process(
+        &matches,
+        &progress,
+        ServeProcessCliOpts::default(),
+        protocol::OpenMode::ReadWrite,
+    )?;
     let mut serve_out = serve_proc.proc.stdout.as_mut().unwrap();
     let mut serve_in = serve_proc.proc.stdin.as_mut().unwrap();
 
@@ -2234,8 +2555,12 @@ fn restore_main(args: Vec<String>) -> Result<(), anyhow::Error> {
     }
 
     let (item_id, query) = cli_to_id_and_query(&matches)?;
-    let mut serve_proc =
-        cli_to_opened_serve_process(&matches, &progress, protocol::OpenMode::Read)?;
+    let mut serve_proc = cli_to_opened_serve_process(
+        &matches,
+        &progress,
+        ServeProcessCliOpts::default(),
+        protocol::OpenMode::Read,
+    )?;
     let mut serve_out = serve_proc.proc.stdout.as_mut().unwrap();
     let mut serve_in = serve_proc.proc.stdin.as_mut().unwrap();
 
@@ -2325,7 +2650,7 @@ fn restore_main(args: Vec<String>) -> Result<(), anyhow::Error> {
             restore_xattrs: matches.opt_present("xattrs"),
         },
         content_index,
-        matches.opt_str("pick"),
+        matches.opt_str("pick").map(|x| x.into()),
         &mut serve_out,
         &mut serve_in,
         &into_dir,
@@ -2354,7 +2679,7 @@ fn serve_main(args: Vec<String>) -> Result<(), anyhow::Error> {
     opts.optflag(
         "",
         "allow-remove",
-        "Allow client to remove repository entries.",
+        "Allow client to remove repository items, implies --allow-list.",
     );
     opts.optflag(
         "",
@@ -2364,12 +2689,17 @@ fn serve_main(args: Vec<String>) -> Result<(), anyhow::Error> {
     opts.optflag(
         "",
         "allow-get",
-        "Allow client to get data from the repository.",
+        "Allow client to retrieve data from the repository, implies --allow-list.",
     );
     opts.optflag(
         "",
         "allow-list",
-        "Allow client to list snapshots and snapshot file lists.",
+        "Allow client to list items and file lists.",
+    );
+    opts.optflag(
+        "",
+        "allow-sync",
+        "Allow client to sync items into the repository, i.e. be the destination of a repository sync.",
     );
 
     let matches = parse_cli_opts(opts, &args[..]);
@@ -2384,6 +2714,7 @@ fn serve_main(args: Vec<String>) -> Result<(), anyhow::Error> {
     let mut allow_gc = true;
     let mut allow_get = true;
     let mut allow_list = true;
+    let mut allow_sync = true;
 
     if matches.opt_present("allow-init")
         || matches.opt_present("allow-put")
@@ -2391,12 +2722,14 @@ fn serve_main(args: Vec<String>) -> Result<(), anyhow::Error> {
         || matches.opt_present("allow-gc")
         || matches.opt_present("allow-get")
         || matches.opt_present("allow-list")
+        || matches.opt_present("allow-sync")
     {
         allow_init = matches.opt_present("allow-init");
-        allow_put = matches.opt_present("allow-put");
         allow_remove = matches.opt_present("allow-remove");
         allow_gc = matches.opt_present("allow-gc");
         allow_get = matches.opt_present("allow-get");
+        allow_sync = matches.opt_present("allow-sync");
+        allow_put = matches.opt_present("allow-put");
         // --allow-get and --allow-remove implies --allow-list because they
         // are essentially useless without being able to list items.
         allow_list = allow_get || allow_remove || matches.opt_present("allow-list");
@@ -2422,6 +2755,7 @@ fn serve_main(args: Vec<String>) -> Result<(), anyhow::Error> {
             allow_gc,
             allow_get,
             allow_list,
+            allow_sync,
             repo_path: Path::new(&matches.free[0]).to_path_buf(),
         },
         &mut stdin_unbuffered,
@@ -2435,6 +2769,76 @@ fn serve_main(args: Vec<String>) -> Result<(), anyhow::Error> {
     result
 }
 
+fn exec_with_locks_main(args: Vec<String>) -> Result<(), anyhow::Error> {
+    let mut opts = default_cli_opts();
+    opts.optopt(
+        "r",
+        "repository",
+        "Repository to lock. \
+         Defaults to BUPSTASH_REPOSITORY if not set. \
+         Unlike other commands, does not support remote repository access.",
+        "REPO",
+    );
+
+    let matches = parse_cli_opts(opts, &args[..]);
+
+    let repo = match matches.opt_str("repository") {
+        Some(repo) => repo,
+        None => match std::env::var("BUPSTASH_REPOSITORY") {
+            Ok(repo) => repo,
+            Err(_) => anyhow::bail!("you must specify either -r or BUPSTASH_REPOSITORY"),
+        },
+    };
+
+    if repo.starts_with("ssh://") {
+        anyhow::bail!("exec-with-locks does not support remote repositories.");
+    }
+
+    if matches.free.is_empty() {
+        die("Expected a command to exec.".to_string());
+    }
+
+    let mut fstx_lock_path = PathBuf::from(repo.clone());
+    fstx_lock_path.push("tx.lock");
+    let mut repo_lock_path = PathBuf::from(repo);
+    repo_lock_path.push("repo.lock");
+
+    // We open files with unsafe so we can disable CLOEXEC, in this
+    // case we explicitly want our child to inherit these locks.
+    let fstx_lock_file = unsafe {
+        std::fs::File::from_raw_fd(nix::fcntl::open(
+            &fstx_lock_path,
+            nix::fcntl::OFlag::O_RDWR,
+            nix::sys::stat::Mode::empty(),
+        )?)
+    };
+    let repo_lock_file = unsafe {
+        std::fs::File::from_raw_fd(nix::fcntl::open(
+            &repo_lock_path,
+            nix::fcntl::OFlag::O_RDWR,
+            nix::sys::stat::Mode::empty(),
+        )?)
+    };
+
+    let fstx_lock = fsutil::FileLock::exclusive_on_file(fstx::FSTX_LOCK_CTX_TAG, fstx_lock_file)?;
+    let repo_lock =
+        fsutil::FileLock::exclusive_on_file(repository::REPO_LOCK_CTX_TAG, repo_lock_file)?;
+
+    let bin = std::ffi::CString::new(matches.free[0].clone()).unwrap();
+    let args: Vec<std::ffi::CString> = matches
+        .free
+        .into_iter()
+        .map(|a| std::ffi::CString::new(a).unwrap())
+        .collect();
+
+    nix::unistd::execvp(&bin, &args)?;
+
+    // Manually drop to ensure these are alive until after exec call.
+    std::mem::drop(fstx_lock);
+    std::mem::drop(repo_lock);
+    anyhow::bail!("exec failed");
+}
+
 fn main() {
     crypto::init();
     cksumvfs::register_cksumvfs();
@@ -2465,6 +2869,8 @@ fn main() {
         "serve" => serve_main(args),
         "recover-removed" => recover_removed(args),
         "put-benchmark" => put_benchmark(args),
+        "sync" => sync_main(args),
+        "exec-with-locks" => exec_with_locks_main(args),
         "version" | "--version" => {
             args[0] = "version".to_string();
             version_main(args)
